/*****************************************************************************
*
* File              : TWI_Master_c.c - vernderte Datei !!
* Compiler          : WINAVR 3.4.6
*
* AppNote           : AVR315 - TWI Master Implementation
*
* Description       : This is a sample driver for the TWI hardware modules.
*                     It is interrupt driveren. All functionality is controlled through
*                     passing information to and from functions. Se main.c for samples
*                     of how to use the driver.
*
* nderung: in ISR(TWI) bei switch / default: -> Stop senden !!
* Zur Unterscheidung mit Slave-Bibliothek sind einige Funktionnamen gendert
* die Union TWI_statusReg ist auf eine 'normale' Variable reduziert
* Der Funktionsaufruf 'TWI_Transceiver_Busy( void )' ist durch den Inhalt der
* Funktion 'while ( TWCR & (1<<TWIE) )' ersetzt.
*
****************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include "TWI_Master_c.h"
#include "global.h"

static uint8_t TWI_MA_buf[TWI_BUF_SIZE];			// Transceiver buffer
//static uint8_t TWI_MA_buf[32];		// Transceiver buffer
static uint8_t TWI_msgSize;                		// Number of bytes to be transmitted.
static uint8_t TWI_state = TWI_NO_STATE;   		// State byte. Default set to TWI_NO_STATE.

static uint8_t TWI_success = TRUE;

// ----------------------------------------------------------------------------
// anstelle der globalen Variablen TWI_statusReg wird eine Funktion zur Abfrage des Bits benutzt
uint8_t TWI_LastTransOK(void)
{
while ( TWCR & (1<<TWIE) );            		// Wait until TWI has completed the transmission.
return(TWI_success);
}

/****************************************************************************
Call this function to set up the TWI master to its initial standby state.
Remember to enable interrupts from the main application after initializing the TWI.
****************************************************************************/
void TWI_Master_Initialise(void)
{
   TWBR = TWI_TWBR;                             // Set bit rate register (Baudrate). Defined in header file.
// TWSR = TWI_TWPS;                             // Not used. Driver presumes prescaler to be 00.
   TWDR = 0xFF;                                 // Default content = SDA released.
   TWCR = (1<<TWEN)|                            // Enable TWI-interface and release TWI pins.
          (0<<TWIE)|(0<<TWINT)|                 // Disable Interupt.
          (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|      // No Signal requests.
          (0<<TWWC);
}

/*
****************************************************************************
Call this function to test if the TWI_ISR is busy transmitting.
****************************************************************************
uint8_t TWI_Transceiver_Busy( void )
{
  return ( TWCR & (1<<TWIE) );                  // IF TWI Interrupt is enabled then the Transceiver is busy
}
*/


/****************************************************************************
Call this function to fetch the state information of the previous operation. The function will hold execution (loop)
until the TWI_ISR has completed with the previous operation. If there was an error, then the function
will return the TWI State code.
****************************************************************************/
/*
uint8_t TWI_Get_State_Info( void )
{
	while ( TWCR & (1<<TWIE) );            		// Wait until TWI has completed the transmission.
	return ( TWI_state );                       // Return error state.
}
*/
/****************************************************************************
Call this function to send a prepared message. The first byte must contain the slave address and the
read/write bit. Consecutive bytes contain the data to be sent, or empty locations for data to be read
from the slave. Also include how many bytes that should be sent/read including the address byte.
The function will hold execution (loop) until the TWI_ISR has completed with the previous operation,
then initialize the next operation and return.
****************************************************************************/
void TWI_MA_Start_Transceiver_With_Data( uint8_t *msg, uint8_t msgSize )
{
   uint8_t temp;

	while ( TWCR & (1<<TWIE) );           			// Wait until TWI is ready for next transmission.

   TWI_msgSize = msgSize;                      	// Number of data to transmit.
   TWI_MA_buf[0]  = msg[0];                    	// Store slave address with R/W setting.
   if (!( msg[0] & (0x01) ))     					// If it is a write operation, then also copy data.
   {
   for ( temp = 1; temp < msgSize; temp++ )
      TWI_MA_buf[ temp ] = msg[ temp ];
   }

	TWI_success = FALSE;
   TWI_state   = TWI_NO_STATE ;
   TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|      // Initiate a START condition.
          (0<<TWWC);
}


/*
// ----------------------------------------------------------------------------
void TWI_MA_write( uint8_t *msg, uint8_t msgSize )
{
   uint8_t temp;

	while ( TWCR & (1<<TWIE) );           			// Wait until TWI is ready for next transmission.
   TWI_msgSize = msgSize;                      	// Number of data to transmit.
   TWI_MA_buf[0]  = (msg[0] & 0xFE);           	// ein gesetztes Read-Bit lschen

   for ( temp = 1; temp < msgSize; temp++ )
      TWI_MA_buf[ temp ] = msg[ temp ];

	TWI_success = FALSE;
   TWI_state   = TWI_NO_STATE ;
   TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|      // Initiate a START condition.
          (0<<TWWC);
}
*/

/*
// ----------------------------------------------------------------------------
void TWI_MA_read( uint8_t *msg, uint8_t msgSize )
{
   uint8_t i;

	while ( TWCR & (1<<TWIE) );           			// Wait until TWI is ready for next transmission.

   TWI_msgSize = msgSize;                      	// Number of data to transmit.
   TWI_MA_buf[0]  = (msg[0] | 0x01);            // Store slave address with R/W setting.
// ######
	TWI_MA_buf[0] |= 0x01;								// das Read-Bit setzen


	TWI_success = FALSE;
   TWI_state         = TWI_NO_STATE ;
   TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|      // Initiate a START condition.
          (0<<TWWC);


	while ( TWCR & (1<<TWIE) );            		// Wait until TWI is ready for next transmission.

	if (TWI_success)            						// Last transmission competed successfully.
   {
   for ( i=0; i<msgSize; i++ )                  // Copy data from Transceiver buffer.
      {
         msg[ i ] = TWI_MA_buf[ i ];
      }
   }
}
*/
/*
****************************************************************************
Call this function to resend the last message. The driver will reuse the data previously put in the transceiver buffers.
The function will hold execution (loop) until the TWI_ISR has completed with the previous operation,
then initialize the next operation and return.
****************************************************************************
void TWI_MA_Start_Transceiver( void )
{
//   while ( TWI_Transceiver_Busy() );            // Wait until TWI is ready for next transmission.
	while ( TWCR & (1<<TWIE) );
   TWI_statusReg.all = 0;
   TWI_state         = TWI_NO_STATE ;
   TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|      // Initiate a START condition.
          (0<<TWWC);
}
*/

/****************************************************************************
Call this function to read out the requested data from the TWI transceiver buffer. I.e. first call
TWI_Start_Transceiver to send a request for data to the slave. Then Run this function to collect the
data when they have arrived. Include a pointer to where to place the data and the number of bytes
requested (including the address field) in the function call. The function will hold execution (loop)
until the TWI_ISR has completed with the previous operation, before reading out the data and returning.
If there was an error in the previous transmission the function will return the TWI error code.
****************************************************************************/
uint8_t TWI_MA_Get_Data_From_Transceiver( uint8_t *msg, uint8_t msgSize )
{
   uint8_t i;

	while ( TWCR & (1<<TWIE) );            		// Wait until TWI is ready for next transmission.
	if (TWI_success)            						// Last transmission competed successfully.
   {
   for ( i=0; i<msgSize; i++ )                  // Copy data from Transceiver buffer.
      {
         msg[ i ] = TWI_MA_buf[ i ];
      }
   }

   return( TWI_success );
}


// ********** Interrupt Handlers ********** //
/****************************************************************************
This function is the Interrupt Service Routine (ISR), and called when the TWI interrupt is triggered;
that is whenever a TWI event has occurred. This function should not be called directly from the main
application.
****************************************************************************/
ISR(TWI_vect)
{
   static uint8_t TWI_bufPtr;

   switch (TWSR)
   {
   case TWI_START:              						// START has been transmitted
   case TWI_REP_START:          						// Repeated START has been transmitted
      TWI_bufPtr = 0;           						// Set buffer pointer to the TWI Address location

// TWI Master Transmitter status codes
   case TWI_MTX_ADR_ACK:        						// SLA+W has been transmitted and ACK received
   case TWI_MTX_DATA_ACK:       						// Data byte has been transmitted and ACK received
      if (TWI_bufPtr < TWI_msgSize)
			{
        TWDR = TWI_MA_buf[TWI_bufPtr++];			// Adresse bzw. Daten nach TWDR schreiben
        TWCR = (1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to send byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|
               (0<<TWWC);
			}
      else                                      // Send STOP after last byte
			{

			TWI_success = TRUE;       					// Set status bits to completed successfully.
			TWCR = (1<<TWEN)|                      // TWI Interface enabled
               (0<<TWIE)|(1<<TWINT)|            // Disable TWI Interrupt and clear the flag
               (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)| // Initiate a STOP condition.
               (0<<TWWC);
			}
      break;

// TWI Master Receiver status codes
   case TWI_MRX_DATA_ACK:                       // Data byte has been received and ACK transmitted
      TWI_MA_buf[TWI_bufPtr++] = TWDR;				// gelesene Daten in den Puffer schreiben

   case TWI_MRX_ADR_ACK:                        // SLA+R has been tramsmitted and ACK received
      if (TWI_bufPtr < (TWI_msgSize-1) )        // Detect the last byte to NACK it.
			{
        TWCR = (1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
               (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send ACK after reception
               (0<<TWWC);
			}
      else                                      // Send NACK after next reception
			{
        TWCR = (1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send NACK after reception
               (0<<TWWC);
			}
      break;

   case TWI_MRX_DATA_NACK:                     // Data byte has been received and NACK transmitted
      TWI_MA_buf[TWI_bufPtr] = TWDR;
		TWI_success = TRUE;         					// Set status bits to completed successfully.
      TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (0<<TWIE)|(1<<TWINT)|              // Disable TWI Interrupt and clear the flag
             (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|   // Initiate a STOP condition.
             (0<<TWWC);
		break;

   case TWI_ARB_LOST:                           // Arbitration lost
      TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|              // Enable TWI Interupt and clear the flag
             (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|   // Initiate a (RE)START condition.
             (0<<TWWC);
      break;
//		alle anderen Flle laufen ber default:
//    case TWI_MTX_ADR_NACK:                    // SLA+W has been transmitted and NACK received
//    case TWI_MRX_ADR_NACK:                    // SLA+R has been transmitted and NACK received
//    case TWI_MTX_DATA_NACK:                   // Data byte has been transmitted and NACK received
//    case TWI_NO_STATE;                        // No relevant state information available; TWINT = 0
//    case TWI_BUS_ERROR:                       // Bus error due to an illegal START or STOP condition
   default:
      TWI_state = TWSR;                         // Store TWSR and automatically sets clears noErrors bit.
                                                // Reset TWI Interface
      TWCR = (1<<TWEN)|                         // Enable TWI-interface and release TWI pins
             (0<<TWIE)|(1<<TWINT)|              // Disable Interupt   - nderung: clear flag
             (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|   // No Signal requests - nderung: Stop Condition senden
             (0<<TWWC);
   }
}


