// --- [frames_c.c ] -----------------------------------------------------------
//
// hier werden empfangene Frames vom Coordinator behandelt

#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "global.h"
#include "xbee_api_c.h"
#include "rs232_int.h"
#include "TWI_Master_c.h"
#include "xbee_api_c.h"


#define EEPROM_TWI_ADR 162
#define ON	1
#define OFF 0

// Position 0 im eingelesenen Array ist die TWI-Adresse, daher bei enum mit 1 beginnen
enum {PIN_MASK = 1, SENDING_XBEE_IDX};
enum {AKTOR_XBEE_IDX = 1, AKTOR_TYP, AKTOR_CMD1, AKTOR_CMD2, AKTOR_OFF, AKTOR_ON, AKTOR_STATE_COUNTER, AKTOR_CONDITION};
enum {OFFSET_AKTOR = 1, OFFSET_SPECIALBITS , OFFSET_CONDITION };
enum {TASTER = 1, ABSOLUT_SCHALTER = 2, PARALLEL_SCHALTER = 4}; // alles andere = Wechselschalter

// --- lokale Variablen -------------------------------------------------------
uint8_t data[64];								// zum Abholen von Daten
char cmd_buffer[64];							// um zu sendende Daten zusammenzustellen

#if TEST_CONDITIONS == TRUE
uint8_t sensor_condition_state[32];		//
#endif
uint8_t sensor_last_pin_state[32];		// hier wird der Status aller 256 Input-Pins gespeichert
uint8_t aktor_last_state[32];				// hier wird der Status aller 256 Aktoren gespeichert
uint8_t aktor_state_counter[4];			// Zhler fr "Parallelschalter"

const char RETURN[] = {13, 0};			// fr die Datenausgabe
const char SPACE[]  = {' ', 0};

// --- lokale Funktionen ------------------------------------------------------

void make_log_header(void);

// ----------------------------------------------------------------------------
// xbee_last_pin_state[32] ist eine Array, in dem fr jeden der 256 XBee_PortPins
// der letzte Pin-Status gespeichert ist.
// Der Sinn ist zu erkennen, ob der Status eines Pins gendert wurde.
//
// hier wird ein Bit an der Position bit_pos gesetzt/gelscht
void sensor_set_last_pin_state(uint8_t bit_pos, uint8_t bit)
{
uint8_t i,j;

j = bit_pos / 8;			// das zustndige Byte aus dem Index ermitteln
i = bit_pos % 8;			// Bitposition

if (bit) sensor_last_pin_state[j] |= (1<<i); else sensor_last_pin_state[j] &= ~(1<<i);
}

// ----------------------------------------------------------------------------
// holt aus dem Array xbee_last_state[32] den letzten Status eines der
// 256 XBee-Portpin
// liefert 0/1 FALSE/TRUEzurck
uint8_t sensor_get_last_pin_state(uint8_t bit_pos)
{
uint8_t i, j, r;

j = bit_pos / 8;			// das zustndige Byte aus dem Index ermitteln
i = bit_pos % 8;			// Bitposition

r = (sensor_last_pin_state[j] & (1<<i));
if (r) r = 1;
return(r);
}

#if TEST_CONDITIONS == TRUE
// ----------------------------------------------------------------------------
// condition_state[32] ist ein Array, in dem fr jede der 256 Bedingungen
// der Status TRUE oder FALSE gespeichert ist.
void sensor_set_condition_state(uint8_t bit_pos, uint8_t bit)
{
uint8_t i,j;

j = bit_pos / 8;			// das zustndige Byte aus dem Index ermitteln
i = bit_pos % 8;			// Bitposition

// speichert mit invertierter Logik: 0 = TRUE, 1 = FALSE
if (!(bit)) sensor_condition_state[j] |= (1<<i); else sensor_condition_state[j] &= ~(1<<i);
}

// ----------------------------------------------------------------------------
// holt aus dem Array sensor_condition[32] den Status einer der
// 256 Bedingungen
// liefert 0/1 FALSE/TRUE zurck
uint8_t sensor_get_condition_state(uint8_t bit_pos)
{
uint8_t i, j, r;

j = bit_pos / 8;			// das zustndige Byte aus dem Index ermitteln
i = bit_pos % 8;			// Bitposition

r = (sensor_condition_state[j] & (1<<i));
// liefert das Bit invertiert zurck: 0 = TRUE, 1 = FALSE
if (r) r = 0; else r = 1;
return(r);
}
#endif

// ----------------------------------------------------------------------------
// aktor_last_state[32] ist ein Array, in dem fr jeden der 256 Aktoren
// der letzte Aktor-Status gespeichert ist.
void aktor_set_last_state(uint8_t bit_pos, uint8_t bit)
{
uint8_t i,j;

j = bit_pos / 8;			// das zustndige Byte aus dem Index ermitteln
i = bit_pos % 8;			// Bitposition

if (bit) aktor_last_state[j] |= (1<<i); else aktor_last_state[j] &= ~(1<<i);
}

// ----------------------------------------------------------------------------
// holt aus dem Array aktor_last_state[32] den letzten Status eines der
// 256 Aktoren
// liefert 0/1 FALSE/TRUE zurck
uint8_t aktor_get_last_state(uint8_t bit_pos)
{
uint8_t i, j, r;

j = bit_pos / 8;			// das zustndige Byte aus dem Index ermitteln
i = bit_pos % 8;			// Bitposition

r = (aktor_last_state[j] & (1<<i));
if (r) r = 1;
return(r);
}


// ----------------------------------------------------------------------------
// es wurde ein IO-Sample empfangen
void FRM0x92(void)
{
	uint8_t sending_xbee, port_state, pin_state, pin_mask, ii, j, m;
	uint8_t sensor_data[33];					// Platz fr die Daten des Sensors -> 32 + 1
	uint8_t aktor_data[9];						// Platz fr die Daten des Aktors -> 8 + 1
	uint8_t last_aktor_state, last_sensor_state, last_state_bit_idx; ;
	
	uint16_t eeprom_adr, aktor_adr;
	
//	make_log_header();							// TWI_Adresse und Zeit an den Beginn von cmd_buffer[] einfgen
	
	cmd_buffer[0] = LOGGER_TWI_ADR;
	cmd_buffer[1] = 0;
	strcat(cmd_buffer, "92");					// Framekennung schreiben	
	strcat(cmd_buffer, SEPARATOR_STR);	

	port_state   = FRM92_get_digital_sample_lowbyte();// Sample der unteren 8 Portpins 
	sending_xbee = FRM_get_sender_idx();
	utoa(sending_xbee, cv, 10);
	
	strcat(cmd_buffer, cv);						// Index des Absenders anfgen
	strcat(cmd_buffer, SEPARATOR_STR);	

	FRM92_get_digital_inputs(&cmd_buffer[strlen(cmd_buffer)]);
	
	strcat(cmd_buffer, RETURN);				// return = neue Zeile in der Logdatei
	
														// Daten an den Logger ausgeben
	TWI_MA_Start_Transceiver_With_Data((uint8_t*)cmd_buffer, strlen(cmd_buffer));

	// Daten analysieren und ggf. schalten
	// Der Idx des absendenden XBees ist bekannt, er steht in der Variablen sending_xbee,
	// Dazu die bis zu 8 Schaltdefinition aus dem Eeprom einlesen.
	// Die StartPosition im Eeprom berechnet sich aus dem Idx des XBee in der Tabelle1 * 32
	// fr jedes der 256 XBees ist ein Speicherbereich von 32 Byte freigehalten,
	// bentutz werden nur 16 Byte:
	// fr 8 Schaltdefinitonen jeweils 2 Byte, 
	// - das 1. Byte legt die PinMaske fest,
	// - das 2. Byte die Adresse der Schalterdefinition in Tabelle 3
	// - das 3. Byte mit Spezialfunktionen: Pin_State invertieren, 
	// - das 4. Byte mit einer Condition, die erfllt sein muss, damit der Sensor aktiv wird
	
	sensor_data[0] = EEPROM_TWI_ADR;			// vorbereiten zum Einlesen der Daten, hier EEprom Adr
	eeprom_adr  = (uint16_t) (sending_xbee * 32); // die Adresse im Eeprom, an der die Zuordnung Pin/Aktor steht
	sensor_data[1] = (uint8_t) (eeprom_adr >> 8);
	sensor_data[2] = (uint8_t)  eeprom_adr;
	TWI_MA_Start_Transceiver_With_Data(sensor_data, 3); // die Startadresse senden

	sensor_data[0]++;								// die Read-Adresse erzeugen
	TWI_MA_Start_Transceiver_With_Data(sensor_data, 33); // die Daten einlesen (+1 Byte fr TWI-Adresse)
	TWI_MA_Get_Data_From_Transceiver(sensor_data, 33); // jetzt stehen die Daten in sensor_data ab Pos. 1

														// Die Position des last_sensor_state im Array xbee_last_pin_state[]
	last_state_bit_idx = (sending_xbee * 8) - 1;
														// berechnet sich aus Sender_idx * 8 + Pinnummer.
														// -1, damit das Incrementieren des Index berschaubar und
														// immer genau am Schleifenbeginn steht
	
	pin_mask = 1;									// Die PinMaske tastet nacheinander alle Bits von port_status ab
													
														// An Pos.0 in sensor_data[] steht die TWI-Adresse !
														// daher mit Index 1 beginnen.
	for(ii = 1; ii < 33; ii += 4)				// die 8 Pin-Aktordefinitionen testen					
	{													// Die Maske darf nur aus einem einzigen Bit bestehen,
		last_state_bit_idx++;					// Fr jeden Schleifendurchgang den Index incrementieren
		

														
		if (sensor_data[ii] == 0xFF) break; // wenn als PinMaske 0xFF eingetragen ist, dann beenden

#if TEST_CONDITIONS == TRUE
		if (!(sensor_get_condition_state(sensor_data[ii + OFFSET_CONDITION])))// wenn die Bedingung nicht erfllt ist
			{

			aktor_data[0] = 4;
			aktor_data[1] = 'C';
			aktor_data[2] = '-';
			TWI_MA_Start_Transceiver_With_Data(aktor_data, 3);		
			continue; 								// dann zum nchsten Datensatz
			}
#endif
														// in sensor_data[ii] steht die Pinmaske, hier darf nur genau
														// ein einziges Bit gesetzt sein 
		m = 1;										// Testen, ob wirklich nur 1 Bit in der PinMaske gesetzt ist:
		j = 8;										// m ist die Maske, 							
		while (j--)
			{
			if((sensor_data[ii] & m) == sensor_data[ii])	// wenn diese Bedingung erfllt ist,
				{										// dann ist nur genau 1 Bit gesetzt.
				m = 255;								// als Zeichen fr den Erfolg 255 eintragen
				break;								// und die weitere Prfung beenden
				}
			m *= 2;									// das Maskierungsbit nach links schieben
			}
			
		if (m == 255)								// dann ist wirklich nur genau 1 Bit gesetzt.
		{												// wenn mehrere Pins als PinChange aktiviert sind, wird
														// nur 1 davon den PinChange ausgelst haben
														// Der letzte Status des Sensors ist in last_state_bit[] gespeichert
														// last_state_bit_idx zeigt auf das zugehrige Bit.
														// Den letzten Status des Pins holen, liefert 0 oder 1
			last_sensor_state = sensor_get_last_pin_state(last_state_bit_idx); 
														// den Port_status mit der Maske in sensor_data[ii] maskieren
														// und als pin_state in das Format 0/1 umwandeln
														// fr den folgenden Vergleich mit last_sensor_state
			if (port_state & sensor_data[ii]) pin_state = 1; else pin_state = 0;

														// Wenn das "Spezialbyte" == 0, dann wird es nicht weiter bercksichtigt
//			if (sensor_data[ii + OFFSET_SPECIALBITS])
//				{										// wenn das Bit.4 im SpezialByte gesetzt ist,
														// dann wird der pin_state invertiert
				if (sensor_data[ii + OFFSET_SPECIALBITS] & 0x10) pin_state ^= 0x01; 
//				}
														// wenn der neue Status ungleich dem letzten Status
														// dann gibt es etwas zu tun - andernfalls zum nchsten Pin wechseln
			if(!(pin_state == last_sensor_state))
			{											// den aktuellen Pinstatus speichern
				sensor_set_last_pin_state(last_state_bit_idx, pin_state);				
														// nun die Daten der Aktortabelle lesen
														// die Daten folgen auf die 32 * 32 Byte fr die Sensoren
														// hier sind pro Aktor zustzliche 16 Bit als Reserve freigehalten
														// Startadresse ist 256 * 8 = 1024 
														// der Offset errechnet sich aus Aktor_idx * 8
														// der Aktor steht in sensor_data[ii + 1]
				aktor_adr = (uint16_t) (1024 + (((uint16_t)sensor_data[ii + OFFSET_AKTOR] * 8))); 
				aktor_data[0] = EEPROM_TWI_ADR;
				aktor_data[1] = (uint8_t) (aktor_adr >> 8);
				aktor_data[2] = (uint8_t)  aktor_adr;
				TWI_MA_Start_Transceiver_With_Data(aktor_data, 3);
						
				aktor_data[0]++;					// die TWI-Adresse auf Read-umschalten
				TWI_MA_Start_Transceiver_With_Data(aktor_data, 9);
				TWI_MA_Get_Data_From_Transceiver(aktor_data, 9); // Nun stehen die Daten in aktor_data
	
														// wenn das Bit.0 oder 1 im Spezialbyte von sensor_data[] gesetzt ist,
														// dann wird der Aktortyp aus aktor_data berschrieben !
														// es darf nur 1 der beiden Bits gesetzt sein, dies beschreibt
														// den Aktortyp -> 1 = Taster, 2 = Absolutschalter, 
				if (sensor_data[ii + OFFSET_SPECIALBITS] & 0x03) aktor_data[AKTOR_TYP] = (sensor_data[ii + OFFSET_SPECIALBITS] & 0x03); 

				// Sonderfall: Wenn der Schalter ein Taster ist und der gerade AUS-geschaltet wurde,
				// dann wird keine Aktion ausgefhrt. Ausgeschaltet = 1 !!.
				if ((aktor_data[AKTOR_TYP] == TASTER) && (pin_state))	continue;
				
				// das Remote AT-CMD an das XBEE schon einmal vorbereiten
				// die beiden Zeichen des Kommandos stehen in aktor_data[] und werden nach xbee_data[] kopiert.
				xbee_data[0] = aktor_data[AKTOR_CMD1];
				xbee_data[1] = aktor_data[AKTOR_CMD2];
				
				// Wie war noch der letzte Status des Aktors ?
				last_aktor_state = aktor_get_last_state(sensor_data[ii + OFFSET_AKTOR]);
				
				// Sonderfall: Parallelschalter
				// Ein Parallelschalter kann aus mehreren parallel geschalteten Schaltern bestehen.
				// Ist mindestens einer der Sensoren EINgeschaltet, dann ist der Aktor ON
				// ist keiner der Sensoren EINgeschaltet, dann ist der Aktor OFF
				if (aktor_data[AKTOR_TYP] == PARALLEL_SCHALTER)
				{
/*					// den Zhler fr Ein-/Ausschaltungen fortschreiben, bei jeder Schalterbettigung
					if (pin_state) aktor_state_counter[aktor_data[AKTOR_STATE_COUNTER]]++; 
					else aktor_state_counter[aktor_data[AKTOR_STATE_COUNTER]]--;

					// pin_state ist 1 und last_state = 0 ->> Einschalten
					if ((pin_state) && (aktor_state_counter[aktor_data[AKTOR_STATE_COUNTER]] == 0))
					{
						xbee_data[2] = aktor_data[AKTOR_ON];
						aktor_set_last_state(aktor_data[AKTOR_XBEE_IDX], ON);
					}
					// pin_state ist 0 und last_state = 1 -> Ausschalten
					else if ((!(pin_state)) && (aktor_state_counter[aktor_data[AKTOR_STATE_COUNTER]] == 1))
					{
						xbee_data[2] = aktor_data[AKTOR_OFF];
						aktor_set_last_state(aktor_data[AKTOR_XBEE_IDX], OFF);
					}
					else return;		// keine nderung auslsen
*/					
				}

				// Sonderfall: Absolutschalter
				// hier wird der Aktor zwingend in den Status von pin_state versetzt
				// Ist dieser Status bereits aktiv, dann gibt es nichts zu tun ....
				else if (aktor_data[AKTOR_TYP] == ABSOLUT_SCHALTER)
					{
					if (!(pin_state == last_aktor_state))
						{
						if (pin_state) 			// den Aktor einschalten
							{
							xbee_data[2] = aktor_data[AKTOR_ON];
							}
						else 
							{
							xbee_data[2] = aktor_data[AKTOR_OFF];
							}
														// den neuen Status speichern
							aktor_set_last_state(sensor_data[ii + OFFSET_AKTOR], pin_state);
						}
					
					else	continue;				// der angeforderte Status ist bereits aktiv
														// dann zurck an den Anfang der Schleife	
					}
			
				// alle anderen Flle wechseln den aktuellen Status des Aktors (Wechsel-Schalter)
				else
					{
					if (last_aktor_state) 		// den Aktor togglen, der letzte Status war TRUE
						{
						xbee_data[2] = aktor_data[AKTOR_OFF];	
						}
					else 								// den Aktor togglen, der letzte Status war FALSE
						{
						xbee_data[2] = aktor_data[AKTOR_ON];
						}
														// den alten Status invertieren und als neuen Status speichern
					last_aktor_state ^= 0x01;
					aktor_set_last_state(sensor_data[ii + OFFSET_AKTOR], last_aktor_state);
					}
				
				// das AT-CMD an das Zieldevice senden
				
				// Der Parameter SEND_NO_RESPONSE fhrt merkwrdigerweise dazu, 
				// dass die Nachricht mit einer sprbaren Verzgerung von ca. 1 Sekunde abgeschickt wird !!
				// Wird die Nachricht direkt 2x nacheinander gesendet, dann entfllt die Verzgerung fr den ersten Job,
				// der zweite wird wieder verzgert abgesendet.
				// Workaround: Abhilfe schafft der Parameter SEND_RESPONSE, 
				// obwohl die Besttigung eigentlich garnicht bentigt und ausgewertet wird ....
				// FRM17_send_remote_at_cmd(xbee_data, 3, aktor_data[AKTOR_XBEE_IDX], SEND_NO_RESPONSE);
				
				FRM17_send_remote_at_cmd(xbee_data, 3, aktor_data[AKTOR_XBEE_IDX], SEND_RESPONSE);
				
aktor_data[0] = LOGGER_TWI_ADR;
aktor_data[1] = aktor_data[AKTOR_XBEE_IDX] + '0';
aktor_data[2] = '>';
aktor_data[3] = xbee_data[0];
aktor_data[4] = xbee_data[1];
aktor_data[5] = xbee_data[2] + '0';

TWI_MA_Start_Transceiver_With_Data(aktor_data, 6);

			}
		}												// if (sensor_data[i]) - wenn eine Maske definiert ist
	}													// for (i = 0; i < 16, i += 2)		
}														// frm0x92



// --- [ eof ] ----------------------------------------------------------------