// --- [ rs232_int.c ] ---------------------------------------------------------

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include "rs232_int.h"

volatile uint8_t rs_buffer[TX_BUFFER_SIZE];		// Ringpuffer zum Senden
volatile uint8_t rs_tail = 0, rs_head = 0;

// ----------------------------------------------------------------------------
//
void rs232_init(void)
{
UBRR0L = (uint8_t) TEILER;									// siehe rs232_int.h
UBRR0H = 0;
UCSR0C = (0<<UMSEL01) | (1<<UCSZ01) | (1<<UCSZ00);
UCSR0B = TX_DISABLE;

HANDSHAKE_INIT;
RX_ENABLE;
}

// ----------------------------------------------------------------------------
// ein einzelnes Zeichen auf COM ausgeben,
// dazu das Byte in die Warteschlange einreihen
// und das erste Byte absenden, sofern keine Transaktion luft
// andernfalls bernimmt diese Aufgabe ISR (USART_TX_vect).
void put_c(uint8_t c)
{

register uint8_t x;
x = rs_head;										// x zeigt auf das nchste freie Byte im Puffer

// verhindert, dass der Puffer von hinten her berschrieben wird
// wartet ggf., bis rs_tail durch Senden eines Bytes incrementiert wurde
while (((x + 1) & TX_BUFFER_MASK) == rs_tail);

rs_buffer[x++] = c;								// c in die Warteschlange einreihen
x &= TX_BUFFER_MASK;								// Pointer-berlauf maskieren
rs_head = x;										// Register zurckschreiben

if (UCSR0B == TX_DISABLE)						// Prfen, ob TX disabled ist
	{
	UCSR0B = TX_ENABLE;							// TX enablen
	x = rs_tail;									// Tail einlesen

	UDR0 = rs_buffer[x++];						// und ein Byte auf die Reise schicken
	x &= TX_BUFFER_MASK;							// Pointer incrementieren, maskieren
	rs_tail = x;									// und speichern
	}
}


/*
// ----------------------------------------------------------------------------
// ein einzelnes Zeichen auf COM ausgeben
void put_c(uint8_t Zeichen)
{
while(!(UCSR0A & (1<<UDRE0)));
UDR0 = Zeichen;
}

*/

// ----------------------------------------------------------------------------
// einen String auf COM ausgeben
void put_s(char *s)
{
while(*s) put_c(*s++);
}

// ----------------------------------------------------------------------------
// gibt einen String aus dem Flash aus
void puts_p(const char *progmem_s)
{
register char c;
while ( (c = pgm_read_byte(progmem_s++)) )
	{
   put_c(c);
	}
}

// ----------------------------------------------------------------------------
// Nachdem ein Byte verschickt worden ist, wird der TXC-Interrupt aufgerufen.
// Es wird geprft, ob weitere Daten im Ringpuffer sind.
// Wenn ja, dann wird das nchste Byte ausgegeben,
// wenn nicht, dann wird die bertragung beendet.
// Der Status von UCSR0B gibt zu erkennen, ob eine bertragung aktiv ist.
ISR (USART_TX_vect)
{

register uint8_t r;								// volatile rs_tail in ein Register laden
r = rs_tail;

if (!(r == rs_head))								// wenn rs_tail == rs_head, dann ist der Puffer leer
	{
	UDR0 = rs_buffer[r++];						// Byte in den Sender schreiben
	r &= TX_BUFFER_MASK;							// Pointer incrementieren und maskieren
	rs_tail = r;									// und zurckschreiben
	}

else													// der Puffer ist leer, bertragung beenden
	{
	UCSR0B = TX_DISABLE;							// TXC-Interrupt disablen
	RX_DISABLE;										// vorerst keine weiteren Daten mehr annehmen
	}
}

// --- [ eof ] ----------------------------------------------------------------
